//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper018.cpp - Implementation of mapper #18 (Jaleco SS8806)
//////////////////////////////////////////////////////////////////////

#include "NesMapper018.h"


NesMapper18::NesMapper18(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *&linkPPU) {
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	// Replace the standard PPU memory object with a VRC4 specific one	
	NesPPU *newPPU = new JalecoPPU();
	newPPU->Init();
	newPPU->VRAM->mbytMirroringType = linkPPU->VRAM->mbytMirroringType;
	newPPU->VRAM->link(this);	
	delete mPpu;
	mPpu = newPPU;
	linkPPU = mPpu;
	mCpu->linkCPU(mPpu, mCpu->mApu, mCpu->mController1, mCpu->mController2, this);

	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();
	
	for (int i = 0; i < 8; i++) { ChrBuf[i] = 0; }
	
	intIrqCount = intIrqLatch = 0;
	blnIrqEnable = false;
}


void NesMapper18::initMemory() {	
	if (bytPRGCount > 0) {
		swapPRGBank16(0x8000, 0);			  // First 16k bank to $8000
		swapPRGBank16(0xC000, bytPRGCount-1); // Last 16k bank to $C000
	}
	PrgBuf[0] = 0;
	PrgBuf[1] = 1;
	PrgBuf[2] = (bytPRGCount << 1) - 2;
}


void NesMapper18::write(Uint16 addr, unsigned char data) {		
	switch (addr) {
	case 0x8000:	// Low 4 bits of 8K PRGROM at $8000
		PrgBuf[0] = PrgBuf[0] & 0xF0 | (data & 0x0F);
		swapPRGBank8(0x8000, PrgBuf[0] % (bytPRGCount << 1));
		break;
	case 0x8001:	// High 4 bits of 8K PRGROM at $8000
		PrgBuf[0] = PrgBuf[0] & 0x0F | ((data & 0x0F) << 4);
		swapPRGBank8(0x8000, PrgBuf[0] % (bytPRGCount << 1));
		break;
	
	case 0x8002:	// Low 4 bits of 8K PRGROM at $A000
		PrgBuf[1] = PrgBuf[1] & 0xF0 | (data & 0x0F);
		swapPRGBank8(0xA000, PrgBuf[1] % (bytPRGCount << 1));
		break;
	case 0x8003:	// High 4 bits of 8K PRGROM at $A000
		PrgBuf[1] = PrgBuf[1] & 0x0F | ((data & 0x0F) << 4);
		swapPRGBank8(0xA000, PrgBuf[1] % (bytPRGCount << 1));
		break;

	case 0x9000:	// Low 4 bits of 8K PRGROM at $C000
		PrgBuf[2] = PrgBuf[2] & 0xF0 | (data & 0x0F);
		swapPRGBank8(0xC000, PrgBuf[2] % (bytPRGCount << 1));
		break;
	case 0x9001:	// High 4 bits of 8K PRGROM at $C000
		PrgBuf[2] = PrgBuf[2] & 0x0F | ((data & 0x0F) << 4);
		swapPRGBank8(0xC000, PrgBuf[2] % (bytPRGCount << 1));
		break;

	case 0xA000:	// Low 4 bits of 1K CHRROM at $0000
		ChrBuf[0] = ChrBuf[0] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x0000, ChrBuf[0] % (bytCHRCount << 3) );
		break;
	case 0xA001:	// High 4 bits of 1K CHRROM at $0000
		ChrBuf[0] = ChrBuf[0] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x0000, ChrBuf[0] % (bytCHRCount << 3) );
		break;
	
	case 0xA002:	// Low 4 bits of 1K CHRROM at $0400
		ChrBuf[1] = ChrBuf[1] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x0400, ChrBuf[1] % (bytCHRCount << 3) );
		break;
	case 0xA003:	// High 4 bits of 1K CHRROM at $0400
		ChrBuf[1] = ChrBuf[1] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x0400, ChrBuf[1] % (bytCHRCount << 3) );
		break;

	case 0xB000:	// Low 4 bits of 1K CHRROM at $0800
		ChrBuf[2] = ChrBuf[2] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x0800, ChrBuf[2] % (bytCHRCount << 3) );
		break;
	case 0xB001:	// High 4 bits of 1K CHRROM at $0800
		ChrBuf[2] = ChrBuf[2] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x0800, ChrBuf[2] % (bytCHRCount << 3) );
		break;
	
	case 0xB002:	// Low 4 bits of 1K CHRROM at $0C00
		ChrBuf[3] = ChrBuf[3] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x0C00, ChrBuf[3] % (bytCHRCount << 3) );
		break;
	case 0xB003:	// High 4 bits of 1K CHRROM at $0C00
		ChrBuf[3] = ChrBuf[3] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x0C00, ChrBuf[3] % (bytCHRCount << 3) );
		break;
	
	case 0xC000:	// Low 4 bits of 1K CHRROM at $1000
		ChrBuf[4] = ChrBuf[4] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x1000, ChrBuf[4] % (bytCHRCount << 3) );
		break;
	case 0xC001:	// High 4 bits of 1K CHRROM at $1000
		ChrBuf[4] = ChrBuf[4] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x1000, ChrBuf[4] % (bytCHRCount << 3) );
		break;
	
	case 0xC002:	// Low 4 bits of 1K CHRROM at $1400
		ChrBuf[5] = ChrBuf[5] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x1400, ChrBuf[5] % (bytCHRCount << 3) );
		break;
	case 0xC003:	// High 4 bits of 1K CHRROM at $1400
		ChrBuf[5] = ChrBuf[5] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x1400, ChrBuf[5] % (bytCHRCount << 3) );
		break;
	
	case 0xD000:	// Low 4 bits of 1K CHRROM at $1800
		ChrBuf[6] = ChrBuf[6] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x1800, ChrBuf[6] % (bytCHRCount << 3) );
		break;
	case 0xD001:	// High 4 bits of 1K CHRROM at $1800
		ChrBuf[6] = ChrBuf[6] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x1800, ChrBuf[6] % (bytCHRCount << 3) );
		break;
	
	case 0xD002:	// Low 4 bits of 1K CHRROM at $1C00
		ChrBuf[7] = ChrBuf[7] & 0xF0 | (data & 0x0F);
		swapCHRBank1(0x1C00, ChrBuf[7] % (bytCHRCount << 3) );
		break;
	case 0xD003:	// High 4 bits of 1K CHRROM at $1C00
		ChrBuf[7] = ChrBuf[7] & 0x0F | ((data & 0x0F) << 4);
		swapCHRBank1(0x1C00, ChrBuf[7] % (bytCHRCount << 3) );
		break;

	case 0xE000:	// Bits 0-3 of IRQ counter
		intIrqLatch = (intIrqLatch & 0xFFF0) | (data & 0x0F);
		break;
	case 0xE001:	// Bits 4-7 of IRQ counter
		intIrqLatch = (intIrqLatch & 0xFF0F) | ((data & 0x0F) << 4);
		break;
	case 0xE002:	// Bits 8-11 of IRQ counter
		intIrqLatch = (intIrqLatch & 0xF0FF) | ((data & 0x0F) << 8);
		break;
	case 0xE003:	// Bits 12-15 of IRQ counter
		intIrqLatch = (intIrqLatch & 0x0FFF) | ((data & 0x0F) << 12);
		break;
	
	case 0xF000:	// Load IRQ counter
		intIrqCount = intIrqLatch;
		break;
	
	case 0xF001:	// Enable/Disable IRQs
		blnIrqEnable = (data & 0x01) ? true : false;
		break;

	case 0xF002:
		switch (data & 0x03) {
		case 0x00:
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;
			break;
		case 0x01:			
			mPpu->VRAM->mbytMirroringType = VERTICAL;
			break;
		case 0x02:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2000;
			break;
		case 0x03:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2400;
			break;
		}
		break;
	}	
}