//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper075.cpp - Implementation of mapper #75 (Jaleco/Konami VRC1)
//////////////////////////////////////////////////////////////////////

#include "NesMapper075.h"


NesMapper75::NesMapper75(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU) {
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();
}


void NesMapper75::initMemory() {	
	if (bytPRGCount > 0) {
		swapPRGBank16(0x8000, 0);				// First 16k bank to $8000
		swapPRGBank16(0xC000, bytPRGCount-1);	// Last 16k bank to $C000
	}
	
	if (bytCHRCount) swapCHRBank8(0);		  // First 8K CHRROM to VRAM

	reg[0] = 0;
	reg[1] = 1;	
}


void NesMapper75::write(Uint16 addr, unsigned char data) {		
	switch (addr & 0xF000) {
	case 0x8000:	// Select 8k RPGROM at $8000
		swapPRGBank8(0x8000, data % (bytPRGCount << 1) );
		break;
	
	case 0x9000:	// Mirroring select, CHRROM select
		if ( data & 0x01 ) {
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;			
		} else {
			mPpu->VRAM->mbytMirroringType = VERTICAL;			 
		}

		// Update CHRROM registers
		reg[0] = (reg[0] & 0x0F) | ((data & 0x02) << 3);
		reg[1] = (reg[1] & 0x0F) | ((data & 0x04) << 2);
		
		swapCHRBank4(0x0000, reg[0] % (bytCHRCount << 1) );
		swapCHRBank4(0x1000, reg[1] % (bytCHRCount << 1) );
		break;
	
	case 0xA000: // Select 8k RPGROM at $A000
		swapPRGBank8(0xA000, data % (bytPRGCount << 1) );
		break;
	
	case 0xC000: // Select 8k RPGROM at $C000
		swapPRGBank8(0xC000, data % (bytPRGCount << 1) );
		break;

	case 0xE000:
		// Select low 4K CHRROM
		reg[0] = (reg[0] & 0x10) | (data & 0x0F);
		swapCHRBank4(0x0000, reg[0] % (bytCHRCount << 1) );		
		break;

	case 0xF000:
		// Select high 4K CHRROM
		reg[1] = (reg[1] & 0x10) | (data & 0x0F);
		swapCHRBank4(0x1000, reg[1] % (bytCHRCount << 1) );
		break;
	}
}



//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the mapper to the file stream
//-------------------------------------------------------------------
void NesMapper75::saveState(FILE *stream) {
	NesBaseMapper::saveState(stream);
	fwrite(&reg[0], sizeof(reg[0]), 1, stream);		// misc
	fwrite(&reg[1], sizeof(reg[1]), 1, stream);		// misc
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the mapper from the file stream
//-------------------------------------------------------------------
void NesMapper75::loadState(FILE *stream) {
	NesBaseMapper::loadState(stream);
	fread(&reg[0], sizeof(reg[0]), 1, stream);		// misc
	fread(&reg[1], sizeof(reg[1]), 1, stream);		// misc
}