//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes PPU.cpp: implementation of the NesPPU class.
//////////////////////////////////////////////////////////////////////

#include "NesPPU.h"


//////////////////////////////////////////////////////////////////////
// Procedure: NesPPU constructor
//////////////////////////////////////////////////////////////////////
// Description: The NesPPU constructor initializes some trivial values
//////////////////////////////////////////////////////////////////////
NesPPU::NesPPU() {	
	display = NULL;
	mintCurScanLine = 0;	
	VRAM = NULL;
	timers.blnOtherIRQ = false;
	
	// Set these values to the NTSC default
	mintNumScanlines = 262;
	mintNumVBlankScanlines = 20;
}

//////////////////////////////////////////////////////////////////////
// Procedure: NesPPU Deconstructor
//////////////////////////////////////////////////////////////////////
// Description:  Deletes the PPU VRAM component
//////////////////////////////////////////////////////////////////////
NesPPU::~NesPPU()
{
	if (VRAM == NULL) delete VRAM;
	VRAM = NULL;
}



//////////////////////////////////////////////////////////////////////
// Procedure: Init
//////////////////////////////////////////////////////////////////////
// Description:  Initializes the display memory, the VRAM object, the 
// palette data, and sets miscellaneous values.
//////////////////////////////////////////////////////////////////////
void NesPPU::Init()
{	
	if (display != NULL) {
		delete display;
		display = NULL;
	}
	
	// Initialize PPU bus
	regs.VRAMAddr = 0;

	// Initialize the current scanline counter
	// After a lot of testing, it seems the best starting point is not at scanline
	// 0 (the start of VBLank), but at scanline 20 (the first dummy scanline)
	mintCurScanLine = 0;
	
	// The display is a 256x240 bitmap
	display = new unsigned char[0xF000 + 1 + 8];
		
	// Initialize the palette, anyone know a better way?
	// Palette info from Chris Covell's FAQ
	mPalette[0x00] = 0x808080; mPalette[0x01] = 0x003DA6; mPalette[0x02] = 0x0012B0; mPalette[0x03] = 0x440096; mPalette[0x04] = 0xA1005E; mPalette[0x05] = 0xC70028; mPalette[0x06] = 0xBA0600; mPalette[0x07] = 0x8C1700;
	mPalette[0x08] = 0x5C2F00; mPalette[0x09] = 0x104500; mPalette[0x0A] = 0x054A00; mPalette[0x0B] = 0x00472E; mPalette[0x0C] = 0x004166; mPalette[0x0D] = 0x000000; mPalette[0x0E] = 0x050505; mPalette[0x0F] = 0x050505;
	mPalette[0x10] = 0xC7C7C7; mPalette[0x11] = 0x0077FF; mPalette[0x12] = 0x2155FF; mPalette[0x13] = 0x8237FA; mPalette[0x14] = 0xEB2FB5; mPalette[0x15] = 0xFF2950; mPalette[0x16] = 0xFF2000; mPalette[0x17] = 0xD63200;
	mPalette[0x18] = 0xC46200; mPalette[0x19] = 0x358000; mPalette[0x1A] = 0x058F00; mPalette[0x1B] = 0x008A55; mPalette[0x1C] = 0x0099CC; mPalette[0x1D] = 0x212121; mPalette[0x1E] = 0x090909; mPalette[0x1F] = 0x090909;
	mPalette[0x20] = 0xFFFFFF; mPalette[0x21] = 0x0FD7FF; mPalette[0x22] = 0x69A2FF; mPalette[0x23] = 0xD480FF; mPalette[0x24] = 0xFF45F3; mPalette[0x25] = 0xFF618B; mPalette[0x26] = 0xFF8833; mPalette[0x27] = 0xFF9C12;
	mPalette[0x28] = 0xFABC20; mPalette[0x29] = 0x9FE30E; mPalette[0x2A] = 0x2BF035; mPalette[0x2B] = 0x0CF0A4; mPalette[0x2C] = 0x05FBFF; mPalette[0x2D] = 0x5E5E5E; mPalette[0x2E] = 0x0D0D0D; mPalette[0x2F] = 0x0D0D0D;	
	mPalette[0x30] = 0xFFFFFF; mPalette[0x31] = 0xA6FCFF; mPalette[0x32] = 0xB3ECFF; mPalette[0x33] = 0xDAABEB; mPalette[0x34] = 0xFFA8F9; mPalette[0x35] = 0xFFABB3; mPalette[0x36] = 0xFFD2B0; mPalette[0x37] = 0xFFEFA6;
	mPalette[0x38] = 0xFFF79C; mPalette[0x39] = 0xD7E895; mPalette[0x3A] = 0xA6EDAF; mPalette[0x3B] = 0xA2F2DA; mPalette[0x3C] = 0x99FFFC; mPalette[0x3D] = 0xDDDDDD; mPalette[0x3E] = 0x111111; mPalette[0x3F] = 0x111111;

	
	// Initialize the PPU RAM (aka the VRAM)	
	if (VRAM == NULL) delete VRAM;

	VRAM = new NesPPUMemory;
	VRAM->init();
	
	regs.VRAMAddr = 0x0;
	regs.bytControlReg1 = 0x0;
	regs.bytControlReg2 = 0x0;
	regs.bytPPUStatus = 0x80;		// Start with VBlank flag raised
	
	regs.bytVRAMIOReg = 0x0;
	regs.tempVRAMAddr = 0x0;

	int ctr = 0;
	while (ctr < 0xFF) {
		SPRRAM[ctr] = 0;
		ctr++;
	}
}



//////////////////////////////////////////////////////////////////////
// Procedure: calcScanLine
//////////////////////////////////////////////////////////////////////
// Description:
// Draws a scanline to the PPU display (which is just a simple 256x256 array of bytes)
// The scanline to be drawn is maintained within the PPU class, as this routine is meant to
// be repeatedly called from the CPU class and I did not want to have the CPU class messing
// around with the current scanline and what not.
//
// The scanline is drawn as follows:  First, the background is drawn.  Then, both layers
// of sprites are drawn, using some tricks to get them to appear properly.  Then,
// the display buffer, which contains palette indices at this point and not
// palette values, is "unrolled" from indices to palette entries.  This is neccesary 
// in order to get several sprite issues to work properly and (hopefully) efficiently.
//////////////////////////////////////////////////////////////////////
void NesPPU::calcScanLine() {	
	// Clear PPU status register bits for 8 sprites per scanline
	//regs.bytPPUStatus &= 0x9F;
	//regs.bytPPUStatus &= 0xDF;

	timers.intCurPPUCycle = 0;

	// Scanlines 0..19 are VBlank; do nothing



	// Scanline 20 is a dummy/pipeline start scanline, pull down some flags
	// but still make the memory fetches	
	if (mintCurScanLine == mintNumVBlankScanlines) {	
		
		// Clear the NES display to the background color
		clearDisplay();
		
		// Perform dummy fetches
		if (regs.bytControlReg2 & 0x18) {
			for (int fetch=0; fetch < 42; fetch++){
				Uint8 dat = VRAM->read(0x2000);
				dat = VRAM->read(0x0000);
			}
		}
	}

	if (mintCurScanLine == mintNumVBlankScanlines+1) {
		// End of VBLank, take down flag
		regs.bytPPUStatus &= 0x7F;

		// Clear the sprite #0 flag
		mblnSpriteFlag = true;				
		regs.bytPPUStatus &= 0xBF;  // doc is unclear about this, but clear spr 0 flag
		regs.bytPPUStatus &= 0xDF;  // clear the 8 sprites flag, too


		// This used to be done on scanline 20, but seemed too soon.
		// Frame start, copy the temp VRAM address to the "real" VRAM address (i.e. copy $2005 to $2006)
		if ( regs.bytControlReg2 & 0x8) regs.VRAMAddr = regs.tempVRAMAddr;
	}

	// Actual drawing code
	if ( (mintCurScanLine > mintNumVBlankScanlines) && (mintCurScanLine <= mintNumVBlankScanlines + 240) ) {
		mintCurScanLine -= (mintNumVBlankScanlines + 1);
		// Start of scanline: copy bits 0-4 and bit 10 from temp addr to "real" vram addr
		if (regs.bytControlReg2 & 0x8) {			
			regs.VRAMAddr = (regs.VRAMAddr & 0xFFE0) | (regs.tempVRAMAddr & 0x001F);	// copy bits 0-5 (horizontal scroll)
			regs.VRAMAddr = (regs.VRAMAddr & 0xFBFF) | (regs.tempVRAMAddr & 0x0400);	// copy bit 10
		}
						
		if (regs.bytControlReg2 & 0x08) {
			drawBackGround();	// Draw background
		} else if (regs.bytControlReg2 & 0x10) {
			// If sprites are enabled but BG is not, still do fetches
			for (int fetch = 0; fetch < 32; fetch++) {
				Uint8 dat = VRAM->read(0x2000);
				dat = VRAM->read(0x2000);
				dat = VRAM->read(0x0000);
				dat = VRAM->read(0x0000);
			}
		}
						
		if (regs.bytControlReg2 & 0x10) {
			drawSprites();		// Draw Sprites
		} else if (regs.bytControlReg2 & 0x08) {
			// If BG is enabled but sprites are not, still do fetches
			for (int fetch = 0; fetch < 8; fetch++) {
				Uint8 dat = VRAM->read(0x2000);
				dat = VRAM->read(0x2000);
				dat = VRAM->read(0x0000);
				dat = VRAM->read(0x0000);
			}
		}


		// "Unroll" graphics from indice values to actual palette values
		int xctr= 0, os = mintCurScanLine * 256;

		while (xctr <= 14) {
			// if BG clipping !(& 0x20) --> dont draw pixel, draw it "0" instead
			// if sprite clipping, then sprite data wont even be written anyways so draw as normal				
			// Go to 14 bc thats the max background can add a 20
			if (xctr < 0x08) {
				if (display[os] & 0x20) {
					// No pixel data, set to bg color
					display[os] = VRAM->palette[ 0 ];
				} else {
					//display[os] = VRAM->palette[display[os]];
					display[os] = VRAM->palette[ display[os] & 0x1F ];
				}
			} else {
				display[os] = VRAM->palette[display[os] & 0x1F];
			}
			os++;
			xctr++;
		}
		while (xctr <= 0xFF) {
			display[os] = VRAM->palette[display[os]];
			os++;
			xctr++;
		}
		mintCurScanLine += (mintNumVBlankScanlines + 1);
	}
	
	// scanline #261 is a dummy scanline that does nothing
	if ( mintCurScanLine == mintNumVBlankScanlines + 240 + 1 ) {
		// VBlank, set the proper status flag (bit 7 of 2002, 2002 = bytPPUStatus)
		//regs.bytPPUStatus |= 0x80;
	}
		
	// Count from 0...NumScanlines
	mintCurScanLine = (mintCurScanLine + 1) % mintNumScanlines;
}




//////////////////////////////////////////////////////////////////////
// Procedure: drawBackground
//////////////////////////////////////////////////////////////////////
// Description: This draws the background tiles onto display graphics
// buffer based on the mintCurScanline data member.  This is what I call
// a heavily "loopy-ized" background drawing routine, as it follows
// loopy's outstanding document very closely.  So if any problems arise
// here, consult that document and try to see where the code goes wrong.
//////////////////////////////////////////////////////////////////////
void NesPPU::drawBackGround()
{			
	// Set the proper background pattern table address
	Uint16 intBackGroundPT;
	if (regs.bytControlReg1 & 0x10)  {
		intBackGroundPT = 0x1000;
	} else	{
		intBackGroundPT = 0x0000;
	}
				
	int intXCtr = 0;
	Uint8 intPTCtr = draw.bytXPixelOffset;
	Uint16 intTemp;
	bool blnFlipped = false;
	Uint8 bytNTData, bytPattern1, bytPattern2, bytUpperTwoBits, bytATData;
	Uint8 bytIndiceOffset = 0x00;		

	while (intXCtr <= 0xFF) {		
		// Fetch the name table data
		intTemp = (regs.VRAMAddr & 0x0FFF) + 0x2000;				

		bytNTData = VRAM->read(intTemp, false);
				
		// Horizontal wrapping/scrolling
		if ( ((intTemp & 0x1F) == 0x1F))
		{
			intTemp &= 0xFFE0;			// Clear first 5 bits			
			intTemp ^= 0x0400;			// toggle bit 10
		} else {		
			intTemp++;
		}						

		// Fetch attribute table data 
		// att table formula =  [ name table  + 0x2000 + 0x3C0   ] + [ (Y tile / 4) * 8           ] + [ X tile / 4                 ]		
		bytATData = VRAM->read( ((regs.VRAMAddr & 0x0C00) + 0x23C0) + ((regs.VRAMAddr & 0x0380) >> 4) + ((regs.VRAMAddr & 0x001F) >> 2));		
				

		// Now extract the correct two bits from the data depending on which quadrant we are in
		// Logic:  Regs.VRAMAddr is the name table indice (0 to 960)  When bit D1 is set, we are
		// on the right half of a four block set (an AT block)  When bit D7 is set, we are on the
		// lower half of a four block set (AT block)
		if (regs.VRAMAddr & 0x02) {			
			// Right half of block
			if ( regs.VRAMAddr & 0x40) {				
				// Lower right half of block
				bytUpperTwoBits = (bytATData & 0xC0) >> 4;
			} else	{
				// Upper right half of block
				bytUpperTwoBits = (bytATData & 0x0C);	
			}
		} else {
			// Left half of block						
			if ( regs.VRAMAddr & 0x40) {				
				// Lower left half of block
				bytUpperTwoBits = (bytATData & 0x30) >> 2;
			} else	{
				// Upper left half of block
				bytUpperTwoBits = (bytATData & 0x3) << 2;	
			}		
		}
		
		// After fetching AT data,  update/increment the bus address
		regs.VRAMAddr = (regs.VRAMAddr & 0xF000) | (intTemp & 0x0FFF);
		
		// Do background clipping
		bytIndiceOffset = 0x00;
		if ( !(regs.bytControlReg2 & 0x2) && (intXCtr <= 7) ) { 						
			bytIndiceOffset = 0x20;		// part of an elaborate scheme to get 
										// clipping + hit collision to work in concert
		} 

		// Now, retrieve the 2 pattern table bytes
		int intPTOffset = intBackGroundPT + (bytNTData * 0x10) + ((mintCurScanLine + ((regs.VRAMAddr & 0x7000) >> 12)) & 0x07);
		//int intPTOffset = intBackGroundPT + (bytNTData * 0x10) + (((regs.VRAMAddr & 0x7000) >> 12));

		bytPattern1 = VRAM->read(intPTOffset);
		bytPattern2 = VRAM->read(intPTOffset + 0x8);
		//bytPattern1 = VRAM->memory[intPTOffset];
		//bytPattern2 = VRAM->memory[intPTOffset + 0x8];

		
		// Unpack the pattern data to the display
		// Two versions are included, a standard C++ version for Linux
		// and an Intel assembly version for Windows (can be about 15% faster)

		#ifdef _LINUX
			// "Same" draw code		
			unsigned char intBitMask = 0x1 << (7 - intPTCtr);				
			while ( (intPTCtr <= 7) && (intXCtr <= 0xFF)) {		
				// This is a 0-15 palette indice formed from the various tables					
				int intIndice = ((bytPattern1 & intBitMask) >> (7 - intPTCtr))  + ((bytPattern2 & intBitMask) >> (7 - intPTCtr) << 1) + bytUpperTwoBits;
				//if (intIndice % 4) display[intXCtr + mintCurScanLine * 256] =  VRAM->read(0x3F00 + intIndice);
				//if (intIndice & 0x03) display[intXCtr + mintCurScanLine * 256] =  VRAM->palette[intIndice];//VRAM->read(0x3F00 + intIndice);
				if (intIndice & 0x03) display[intXCtr + mintCurScanLine * 256] =  intIndice + bytIndiceOffset;
					
				intBitMask = intBitMask >> 1;
				intXCtr++;
				intPTCtr++;
			}
		#endif
				

		#ifdef _WINDOWS
			/////////////////
			//int pat1 = ((bytPattern1 & 0x0F) << 4) | (bytPattern1 >> 4);
			int pat1 = bytPattern1;
			int pat2 = bytPattern2;
			//int pat2 = ((bytPattern2 & 0x0F) << 4) | (bytPattern2 >> 4);
			int attbits = bytUpperTwoBits;
			Uint8 patend = 0x07;
			if (intXCtr > 248)	patend = 7 - ((intXCtr + 7) - 255);
			
			Uint8 patstart = (unsigned char) (intPTCtr & 0x7);
			unsigned char *memptr = &display[intXCtr + mintCurScanLine * 256];		
			if ((patend - patstart) == 7) {
				__asm {
					mov edi, memptr
					mov eax, pat1	// unpack pattern 1 into edx (all 8 bits)
					mov ecx, 0x11111111
					xor edx, edx
					shl eax, 15		// 3rd byte of pat1
					or edx, eax
					shr eax, 9		// 2nd byte
					or edx, eax
					ror eax, 9
					or edx, eax		// 1st byte 
					ror eax, 5
					or edx, eax		// 4th byte.
					and edx, ecx // mask out all bad bits; now pattern 1 is unpacked and ready to go in edx

					mov eax, pat2   // unpack pattern2 into ebx
					xor ebx, ebx
					shl eax, 15		// 3rd byte of pat1
					or ebx, eax
					shr eax, 9		// 2nd byte
					or ebx, eax
					ror eax, 9
					or ebx, eax		// 1st byte 
					ror eax, 5
					or ebx, eax		// 4th byte.
					and ebx, ecx // mask out all bad bits
					
					shl ebx, 1			// move all bits over 1 spot for combination
					or edx, ebx			// combine pattern table data into edx

					mov eax, attbits	// fill in attribute table bits into ecx
					mov ecx, eax
					shl eax, 8
					or ecx, eax
					shl eax, 8
					or ecx, eax
					shl eax, 8
					or ecx, eax
					mov eax, ecx
									
					// prepare first byte
					ror edx, 4
					or eax, edx			// or in data
					and eax, 0x0F0F0F0F
					stosd

					rol edx, 4
					and eax, 0x0C0C0C0C
					or eax, edx
					and eax, 0x0F0F0F0F
					stosd
				}
				
				intXCtr += (patend - patstart + 1);
			} else {
				unsigned char intBitMask = 0x1 << (7 - intPTCtr);				
				while ( (intPTCtr <= 7) && (intXCtr <= 0xFF)) {		
					// This is a 0-15 palette indice formed from the various tables					
					int intIndice = ((bytPattern1 & intBitMask) >> (7 - intPTCtr))  + ((bytPattern2 & intBitMask) >> (7 - intPTCtr) << 1) + bytUpperTwoBits;
					//if (intIndice % 4) display[intXCtr + mintCurScanLine * 256] =  VRAM->read(0x3F00 + intIndice);
					//if (intIndice & 0x03) display[intXCtr + mintCurScanLine * 256] =  VRAM->palette[intIndice];//VRAM->read(0x3F00 + intIndice);
					if (intIndice & 0x03) display[intXCtr + mintCurScanLine * 256] =  intIndice + bytIndiceOffset;
						
					intBitMask = intBitMask >> 1;
					intXCtr++;
					intPTCtr++;
				}
			}
			/*
				// Clipped routine goes here								
				__asm {
					push edi
						
					cld					// increment upwards instead of downwards on stosXX
						
					xor ecx, ecx		// Clear out ecx
					mov edi, memptr		// EDI = pointer to memory offset
						
					mov cl, patstart
					//add edi, ecx		// Skip to start of pattern (if this is commented, dont need to xor ecx)

					mov ebx, 0x8080
					shr ebx, cl			// EBX = bit ANDer		

					mov al, bytPattern1		// AX = pattern data
					mov ah, bytPattern2		// Append the upper two bits to the left of ah
					mov ch, patend
					mov cl, 0x07
					sub cl, ch
					shr ax, cl				// Shift over to ignore bits on rightend of stream that are clipped (tricky)
					shr ebx, cl

					mov cl, patend		
					sub cl, patstart		// CL = loop counter (CH is free to use)
																								
					mov edx, eax		// EDX = copy of pattern data shifted over, upper two bits in DL
					shl edx, 16
					or dl, bytUpperTwoBits
					or dh, bytIndiceOffset
						
			patternloop:		
					and eax, ebx		// Pull out the two bits we want from pattern
						
					shr ax, cl			// Shift both bits over to the ones position
					shl ah, 1			// Shift the high bit over one spot

					or al, ah			// combine two pattern bits into al
					jz skip
					or al, dl			// add upper two bits			
					or al, dh			// hack 
								
					stosb				// Write byte to memory buffer, auto-incrementing edi			
			resume:			
					mov eax, edx		// Restore pattern data
					shr eax, 16
						
					shr ebx, 1			// shift bit ANDer
					dec cl				// Decrement the counter until we reach the desired limit			
					jns patternloop
					jmp endproc
			skip:
					inc edi
					jmp resume
			endproc:			
					pop edi
				
			}

			intXCtr += (patend - patstart + 1);*/
			/////////////
		#endif

		intPTCtr = 0;					
	}	

	int *ptr= (int*)&display[mintCurScanLine*256];
	for (int i=0; i < 64; i++) {
		int d = ptr[i];
		if ( !(regs.bytControlReg2 & 0x2) && (i < 2) ) d &= 0x20202020;
		if (!(d & 0x03)) d &= 0xFFFFFF00;
		if (!(d & 0x0300)) d &= 0xFFFF00FF;
		if (!(d & 0x030000)) d &= 0xFF00FFFF;
		if (!(d & 0x03000000)) d &= 0x00FFFFFF;
		ptr[i] = d;
	}


	int bsCtr =0 ;
	intTemp += 0x20;
	int intExtraReads = 1 + (draw.bytXPixelOffset == 0);	
	while (bsCtr < intExtraReads) {
		Uint8 gar = VRAM->read(intTemp, false);
		int intPTOffset = intBackGroundPT + (gar * 0x10) + (((mintCurScanLine+1) + ((regs.VRAMAddr & 0x7000) >> 12)) & 0x07);
		//int intPTOffset = intBackGroundPT + (bytNTData * 0x10) + (((regs.VRAMAddr & 0x7000) >> 12));

		Uint8 p1 = VRAM->read(intPTOffset);
		Uint8 p2 = VRAM->read(intPTOffset + 0x8);
		// Horizontal wrapping/scrolling
		if ( ((intTemp & 0x1F) == 0x1F)) {
			intTemp &= 0xFFE0;			// Clear first 5 bits			
			intTemp ^= 0x0400;			// toggle bit 10
		} else {		
			intTemp++;
		}						

		bsCtr++;
	}
	
	intTemp = regs.VRAMAddr;

	// Stay on the current name table row for 8 scanlines
	if ( ((mintCurScanLine + ((regs.VRAMAddr & 0x7000) >> 12) & 0x7) == 7) && (mintCurScanLine != 239)) {	
		intTemp += 0x20;								
	}			

	// Do vertical wrapping - when bits 5-9 reach 29, bit 11 is toggled
	if ( ((intTemp & 0x03E0) >> 5) == 30 )
	{
		intTemp ^= 0x0800;	// toggle bit 11
		intTemp &= 0xFC1F;	// Clear bits 5-9
		regs.VRAMAddr = intTemp;
	}	else if ( (intTemp & 0x03E0) == 0 )	{
		// In this case, do not toggle bit 11 or accidentally add to VRAMAddr
		// (simply zero out bits 5-9)
		regs.VRAMAddr = (regs.VRAMAddr & 0xFC1F);
	} else {		
		regs.VRAMAddr = intTemp;
	}
}






//////////////////////////////////////////////////////////////////////
// Procedure: Draw Sprites
//////////////////////////////////////////////////////////////////////
// Description: This procedure draws the sprite indices to the display buffer.
// The steps for drawing sprites must follow a very peculiar set of
// rules, which accounts for so much code.  Here are some of the biggest
// snags I have hit:
// 1) Sprites may be clipped in the left 8 pixel column
// 2) Clipped sprites still should generate the hit #0 flag
// 3) Only one sprite can occupy a single pixel (thus the "scanline" bool buffer)
// 4) Given point 3, sprites should be drawn from 0 to 64 and not the other way
// 6) 8x16 sprites that have odd addresses should actually subtract one
//    from the given address before fetching data so they align on an even address
//
// All those issues and more are addressed in this procedure...so beware
//
// Written 9.6.02
// Modified 9.12.02 to add sprite #0 hit detection for background sprite #0s
// Modified 9.27.02 to add proper 8x16 sprite drawing
// Modified 11.12.02 to draw all sprites in one pass (background+foreground)
// Modified 12.5.02 to correct priority problems
//////////////////////////////////////////////////////////////////////
void NesPPU::drawSprites() {	
	int intXLoc, intFetches = 0;
	bool scanline[256];
	int temp=0;	while (temp <= 255) {	scanline[temp] = true;	temp++;	}

	// Set the proper sprite pattern table
	int intSpritePatternTable;	
	if (regs.bytControlReg1 & 0x8) {
		intSpritePatternTable = 0x1000;
	} else {
		intSpritePatternTable = 0x0000;
	}

	int intSpriteSize = 8;
	if (regs.bytControlReg1 & 0x20) intSpriteSize = 16;
		
	// Count the number of displayed sprites
	Uint8 bytNumSprites = 0;

	// Loop through the sprites 0-63 and draw the ones with matching background priority
	// (Loop so #0 is drawn last, 63 first, etc)	
	//int ctr = 252;
	bool Spr0Hit = false;
	int ctr = 0;
	//while (ctr >= 0)			
	while (ctr <= 252) {		
		// Increment Y value for performing comparisons (important to use larger data type here)
		int ySprite = SPRRAM[ctr] + 1;

		// Only display 8 sprites per scanline, special exception for sprite #0 (so hit detection code is always run)		
		if ( (ySprite <= 239) && ((bytNumSprites <= 7) || (ctr == 0)) ) {
			Uint8 bytSprPriority = 0;
			if (SPRRAM[ctr + 2] & 0x20) bytSprPriority = 1;
									
			// Check if the sprite falls on the current scanline			
			if ( (mintCurScanLine >= ySprite) && (mintCurScanLine < (ySprite + intSpriteSize)) ) {			
				bytNumSprites++;

				// Set PPU flag for more than 8 sprites per scanline hit				
				if (bytNumSprites == 8) regs.bytPPUStatus |= 0x20;
								
				int intTileIndex = SPRRAM[ctr + 1];
					
				// Change pattern table address if the sprite size is 8x16
				if (intSpriteSize == 16) {
					if (intTileIndex & 0x01) {
						// Odd tile #, use $1000
						intSpritePatternTable = 0x1000;
							
						// Apparently you still have to start with an even number index
						// (figured this one out the hard way, believe me)
						intTileIndex--;
					} else {
						// Even tile #, use $0000
						intSpritePatternTable = 0x0000;
					}
				}
					
				int intPatternYOffset, intTileOffset;
					
				// Take care of Y sprite flipping
				if (SPRRAM[ctr + 2] & 0x80)
				{
					// Vertical flipping, change offset											
					intPatternYOffset = (intSpriteSize - 1 - ((mintCurScanLine - ySprite))) & 0x7;
				} else	{					
					// standard offset
					intPatternYOffset = (mintCurScanLine - ySprite) & 0x7;
				}

				// Calculate the pattern table indice
				intTileOffset = intSpritePatternTable + (intTileIndex * 0x10) + intPatternYOffset;
										
				// Advance to second pattern if needed, for 8x16 sprites
				if (intSpriteSize == 16) {
					if (SPRRAM[ctr + 2] & 0x80) {
						// Flipped, in first pattern, advance to next pattern
						if ((mintCurScanLine - ySprite) <= 7) intTileOffset += 0x10;														
					} else {
						// Not flipped, past first pattern, advance to next pattern
						if ((mintCurScanLine - ySprite) > 7) intTileOffset += 0x10;
					}
				}
						
				// Extract the two pattern table bytes for this scanline				
				// Note: a garbage NT fetch is added here (for A13 toggling) but 
				// 2 garbage fetches are actually performed in the actual NES 
				// pipeline
				Uint8 garbage = VRAM->read( 0x2000 );
				Uint8 bytPattern1 = VRAM->read( intTileOffset );
				Uint8 bytPattern2 = VRAM->read( intTileOffset + 0x8 );
				intFetches++;

				// Get the X co-ordinate of sprite from SPRRAM
				int intXOffset = SPRRAM[ctr + 3];
								
				// Horizontally flip sprite data if neccesary
				if (SPRRAM[ctr + 2] & 0x40)
				{
					Uint8 temp1 = 0, temp2 = 0;
					int intBitMaskSrc = 0x80, intBitMaskDest = 0x01;
					while (intBitMaskSrc >= 0x01)
					{
						if (bytPattern1 & intBitMaskSrc) temp1 |= intBitMaskDest;
						if (bytPattern2 & intBitMaskSrc) temp2 |= intBitMaskDest;
						intBitMaskSrc = intBitMaskSrc >> 1;
						intBitMaskDest = intBitMaskDest << 1;							
					}
					bytPattern1 = temp1;
					bytPattern2 = temp2;
				}
					
				// Do Sprite clipping (explicit (control reg #2 D2) and implicit (past column 255)
				unsigned char bytLen = 7;
				if (SPRRAM[ctr + 3] > 248) bytLen = 255 - SPRRAM[ctr + 3];										
																
				int intBitMask = 0x80, intPatternTableCtr = 0, intIndice;
				Uint8 bytIndiceOffset = 0x10;
				bool blnClipping = !(regs.bytControlReg2 & 0x4);
			
				// Set the X pixel counter for usage with the scanline array (optimization)
				intXLoc = intXOffset;
				
				if (ctr == 0) {					
					// Separate drawing code for sprite #0 with hit detection included
					// Extract and "draw" pattern table data
					while (intPatternTableCtr <= bytLen) {														
						if (scanline[intXLoc]) {
							// This is a 0-15 palette indice formed from the pattern table + SPRRAM color bits
							intIndice = ((bytPattern1 & intBitMask) >> (7 - intPatternTableCtr))  + ((bytPattern2 & intBitMask) >> (7 - intPatternTableCtr) << 1) + ((SPRRAM[ctr + 2] & 0x03) << 2); 
								
							// If non-transparent...
							if (intIndice & 0x03)  {															
								int intMemVal = intXOffset + intPatternTableCtr + mintCurScanLine * 256;
								if ((blnClipping) && ((intXOffset + intPatternTableCtr) <= 0x07) ) {
									// We are clipping sprite #0, so don't even bother writing pixel vals
									if ((display[intMemVal] & 0x1F) && !Spr0Hit) {
										//regs.bytPPUStatus |= 0x40;
										timers.intSpr0HitSL = mintCurScanLine;
										timers.intSpr0HitPix = intXLoc;
										Spr0Hit = true;
									}
								} else {
									// Check the display first for non-transparent pixel, Set hit flag (D6 of status register)
									if (display[intMemVal] & 0x03) {					
										if (!Spr0Hit) {
											//regs.bytPPUStatus |= 0x40;
											timers.intSpr0HitSL = mintCurScanLine;
											timers.intSpr0HitPix = intXLoc;
											Spr0Hit = true;
										}
										if (bytSprPriority == 0) {											
											display[intMemVal] = (display[intMemVal] & 0x20) | (intIndice + bytIndiceOffset);
										} else {
											// dont draw the background sprite pixel, but mark the spot as being drawn											
										}
										scanline[intXLoc] = false;
									} else {																				
										// Sprite pixel is on a transparent background pixel, overwrite data
										display[intMemVal] = (display[intMemVal] & 0x20) | (intIndice + bytIndiceOffset);
										scanline[intXLoc] = false;										
									}
								}							
							}
						}
						intBitMask = intBitMask >> 1;
						intPatternTableCtr++;
						intXLoc++;
					}
				} else	{	
					// Standard sprite drawing code
					// Extract and "draw" pattern table data
					int intDisplayOffset = intXOffset + mintCurScanLine * 256;					
					
					while (intPatternTableCtr <= bytLen)
					{		
						if (scanline[intXLoc]) // if a sprite has not previously been drawn on this pixel
						{
							if ( (blnClipping) && ((intXOffset + intPatternTableCtr) > 0x07) || (!blnClipping)) {
								// This is a 0-15 palette indice formed from the patte table + SPRRAM color bits
								intIndice = ((bytPattern1 & intBitMask) >> (7 - intPatternTableCtr))  | ((bytPattern2 & intBitMask) >> (7 - intPatternTableCtr) << 1) | ((SPRRAM[ctr + 2] & 0x03) << 2); 
																					
								if (intIndice & 0x03) // if not a transparent color
								{
									int intMemVal = intDisplayOffset + intPatternTableCtr;
									if (display[intMemVal])	// if background pixel exists here
									{
										if (!bytSprPriority) // if sprite is in front 
										{																	
											display[intMemVal] = (display[intMemVal] & 0x20) | (intIndice + bytIndiceOffset);
											scanline[intXLoc] = false;																		
										} else {
											// dont draw the background sprite pixel, but mark the spot as being drawn
											scanline[intXLoc] = false;
										}
									} else { // sprite is in back, no background data drawn								
										display[intMemVal] = (display[intMemVal] & 0x20) | (intIndice + bytIndiceOffset);
										scanline[intXLoc] = false;								
									}								
								}
							}
						}												
						intBitMask = intBitMask >> 1;
						intPatternTableCtr++;
						intXLoc++;
					}
				}							
			}
		}
				
		ctr += 4;	// each SPRRAM entry is 4 bytes
	}

	// Make sure we fetch 8 tiles from memory no matter what; these are dummy fetches
	// but might have ot be changed later to something more specific	
	Uint8 data;
	while (intFetches < 8) {
		data = VRAM->read( 0x2000 );	// NT region
		data = VRAM->read( 0x0000 );	// PT region
		intFetches++;
	}
}





//////////////////////////////////////////////////////////////////////
// Procedure: clearDisplay
//////////////////////////////////////////////////////////////////////
// Description:  This paints the entire graphics display the color
// stored as palette entry #0 in the PPU RAM
// In retrospect, a separate function was probably not needed for this
//////////////////////////////////////////////////////////////////////
void NesPPU::clearDisplay() {
	memset(display, 0x00, 0xF000);
}




//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the PPU to the file stream
//-------------------------------------------------------------------
void NesPPU::saveState(FILE *stream) {
	fwrite(&SPRRAM[0], sizeof(SPRRAM[0]) * 0x100, 1, stream);		// Sprite RAM
	fwrite(&regs, sizeof(PPU_regs), 1, stream);  // PPU registers/latches
	fwrite(&mintCurScanLine, sizeof(mintCurScanLine), 1, stream);	// Misc
	fwrite(&mblnSpriteFlag, sizeof(mblnSpriteFlag), 1, stream);		// Misc
	fwrite(&draw, sizeof(draw), 1, stream);							// Misc
	fwrite(&timers, sizeof(timers), 1, stream);						// Misc
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the PPU from the file stream
//-------------------------------------------------------------------
void NesPPU::loadState(FILE *stream) {
	fread(&SPRRAM[0], sizeof(SPRRAM[0]) * 0x100, 1, stream);		// Sprite RAM
	fread(&regs, sizeof(PPU_regs), 1, stream);  // PPU registers/latches
	fread(&mintCurScanLine, sizeof(mintCurScanLine), 1, stream);	// Misc
	fread(&mblnSpriteFlag, sizeof(mblnSpriteFlag), 1, stream);		// Misc
	fread(&draw, sizeof(draw), 1, stream);							// Misc
	fread(&timers, sizeof(timers), 1, stream);						// Misc
}