// Implementation of VRC6 chip
// Most of this was written from Kevin Horton's excellent document

#include "VRC6.h"


VRC6::VRC6() {	
	reset();
}

void VRC6::exec(int intCycles) {
	for (int i = 0; i < intCycles; i++) {
		// Run pulse 1
		if (pulse1.intChannelEnable) {
			// Normal pulse operation
			if (!pulse1.intGateType) {
				if (!(--pulse1.intCycleCtr)) {
					// Update sample
					if (pulse1.intDutyCtr < (pulse1.intDutyType + 1)) {
						pulse1.intCurSample = pulse1.intVolume;
					} else{
						pulse1.intCurSample = 0;
					}
					pulse1.intDutyCtr = (pulse1.intDutyCtr + 1) & 0x0F;

					pulse1.intCycleCtr = pulse1.intFrequency + 1;
				}
			} else {
				// Digitized mode
				pulse1.intCurSample = pulse1.intVolume;
			}
		}

		// Run pulse 2
		if (pulse2.intChannelEnable) {
			// Normal pulse operation
			if (!pulse2.intGateType) {
				if (!(--pulse2.intCycleCtr)) {
					// Update sample
					if (pulse2.intDutyCtr < (pulse2.intDutyType + 1)) {
						pulse2.intCurSample = pulse2.intVolume;
					} else{
						pulse2.intCurSample = 0;
					}
					pulse2.intDutyCtr = (pulse2.intDutyCtr + 1) & 0x0F;

					pulse2.intCycleCtr = pulse2.intFrequency + 1;
				}
			} else {
				// Digitized mode
				pulse2.intCurSample = pulse2.intVolume;
			}
		}

		// Run sawtooth channel
		if (sawtooth.intChannelEnable) {
			if (!(--sawtooth.intCycleCtr)) {
				// Double accumulator
				if (sawtooth.intAccumCtr) sawtooth.intAccum = sawtooth.intAccum + sawtooth.intPhase;
				
				// Increment counter
				sawtooth.intAccumCtr++;
				if (sawtooth.intAccumCtr == 7) {
					sawtooth.intAccum = 0;
					sawtooth.intAccumCtr = 0;
				}
				
				sawtooth.intCycleCtr = sawtooth.intFrequency + 1;
			}
			sawtooth.intCurSample = (sawtooth.intAccum & 0xF8) >> 3;
		} else {
			sawtooth.intCurSample = 0;
		}
	}

	// At end of execution, add samples together
	// (pulses should be 4 bit, sawtooth 5 bit for total of 6 bit sample)
	intCurSample = pulse1.intCurSample + pulse2.intCurSample + sawtooth.intCurSample;	
	//intCurSample = pulse1.intCurSample + pulse2.intCurSample;	
}

void VRC6::write(Uint16 addr, unsigned char data) {
	switch (addr) {
	case 0x9000:	// Duty cycle, volume, gate
		pulse1.intVolume = (data & 0x0F);
		pulse1.intDutyType = (data & 0x70) >> 4;		
		pulse1.intGateType = (data & 0x80);
		break;

	case 0x9001:	// Frequency (lower 8 bits)
		pulse1.intFrequency = (pulse1.intFrequency & 0xF00) | data;
		break;

	case 0x9002:	// Frequency (upper 4 bits), channel disable
		pulse1.intFrequency = (pulse1.intFrequency & 0x0FF) | ((data & 0x0F) << 8);
		pulse1.intChannelEnable = data & 0x80;
		break;
	
	case 0xA000:
		pulse2.intVolume = (data & 0x0F);
		pulse2.intDutyType = (data & 0x70) >> 4;		
		pulse2.intGateType = (data & 0x80);	
		break;

	case 0xA001:
		pulse2.intFrequency = (pulse2.intFrequency & 0xF00) | data;
		break;

	case 0xA002:
		pulse2.intFrequency = (pulse2.intFrequency & 0x0FF) | ((data & 0x0F) << 8);
		pulse2.intChannelEnable = data & 0x80;
		break;
	

	case 0xB000:	// Phase accumulator input
		sawtooth.intPhase = data & 0x3F;
		break;

	case 0xB001:	// Lower 8 bits of frequency
		sawtooth.intFrequency = (sawtooth.intFrequency & 0xF00) | data;
		break;

	case 0xB002:	// Upper 4 bits of frequency, channel disable
		sawtooth.intFrequency = (sawtooth.intFrequency & 0x0FF) | ((data & 0x0F) << 8);
		sawtooth.intChannelEnable = data & 0x80;
		break;
	}
}



void VRC6::reset() {
	memset(&pulse1, 0x00, sizeof(pulse1));
	memset(&pulse2, 0x00, sizeof(pulse2));
	memset(&sawtooth, 0x00, sizeof(sawtooth));

	pulse1.intCycleCtr = pulse2.intCycleCtr = sawtooth.intCycleCtr = 1;
	intCurSample = 0;
}


void VRC6::loadState(FILE *stream) {
	fread( &pulse1, sizeof(pulse1), 1, stream);
	fread( &pulse2, sizeof(pulse2), 1, stream);
	fread( &sawtooth, sizeof(sawtooth), 1, stream);
}


void VRC6::saveState(FILE *stream) {
	fwrite( &pulse1, sizeof(pulse1), 1, stream);
	fwrite( &pulse2, sizeof(pulse2), 1, stream);
	fwrite( &sawtooth, sizeof(sawtooth), 1, stream);
}